from ._sign_cffi_maker import make_sign_ffi
from textwrap import dedent
import re

def make_sphincs_ffi(build_root):
	_hash_name = re.match(r'.*sphincs-(\w+)', str(build_root)).group(1)
	_hash_src = {'sha2': 'sha2.c', 'shake': 'fips202.c'}[_hash_name]
	common_sources = ['randombytes.c', _hash_src]

	return make_sign_ffi(build_root=build_root, common_sources=common_sources)



from ._kem_cffi_maker import make_kem_ffi
from textwrap import dedent

def make_hqc_ffi(build_root):
	common_sources = ['fips202.c', 'randombytes.c']

	patent_info = (
		3, [
		'FR2956541B1/US9094189B2/EP2537284B1',], [
		'https://csrc.nist.gov/csrc/media/Projects/post-quantum-cryptography/documents/round-4/final-ip-statements/HQC-Statements-Round4.pdf']
	)

	return make_kem_ffi(build_root=build_root, common_sources=common_sources, patent_info=patent_info)



from .kem import mceliece6960119

if __name__ == '__main__':
	public_key, secret_key = mceliece6960119.kem_keypair()
	test_key, test_ciphertext = mceliece6960119.kem_enc(public_key)
	test_decrypted = mceliece6960119.kem_dec(test_ciphertext, secret_key)

	if test_key != test_decrypted:
		raise AssertionError("fail :(")
	print("OK")



from cffi import FFI

from distutils.sysconfig import parse_makefile
from pathlib import Path
import platform
import re
import warnings

from pqc._util import partition_list, map_immed, extant_with_other_suffix, patent_warning

_NAMESPACE_RE = re.compile(r'(?ms)^#define\s+(CRYPTO_NAMESPACE)\s*\(\s*(\w+)\s*\)\s+(\w+)\s*##\s*\2\s*$')

def make_pqclean_ffi(build_root, c_header_sources, cdefs, *,
    common_sources=frozenset(),
    parent_module='pqc._lib',
    patent_info=None):

	# 0. local variables #

	build_root = Path(build_root)
	makefile_parsed = parse_makefile(build_root / 'Makefile')
	common_dir = build_root / '..' / '..' / '..' / 'common'
	_lib_name = Path(makefile_parsed['LIB']).stem
	lib_name = _lib_name.replace('-', '_')

	# 1. module_name #

	module_name = f'{parent_module}.{lib_name}'
	if patent_info is not None:
		patent_message = patent_warning(lib_name, patent_info)
		warnings.warn(patent_message)

	# 2. cdefs, c_header_sources #

	m = _NAMESPACE_RE.search((build_root / 'params.h').read_text() if (build_root / 'params.h').exists() else '')
	if m:
		namespace = m.group(3)
	else:
		##warnings.warn(f'falling back to alternate codepath to figure out CRYPTO_NAMESPACE while building {lib_name} from {build_root}')
		m = re.search(r'(?ms)^#define (\w+)CRYPTO_ALGNAME ', (build_root / 'api.h').read_text())
		if m:
			namespace = m.group(1)
		else:
			raise Exception(f"couldn't figure out CRYPTO_NAMESPACE while building {lib_name} from {build_root}")

	# FIXME: I am certain this isn't the correct function for this
	cdefs = [s.replace('%(namespace)s', namespace) for s in cdefs]
	c_header_sources = [s.replace('%(namespace)s', namespace) for s in c_header_sources]

	# 3. sources, extra_objects #

	if 'SOURCES' in makefile_parsed.keys():
		_source_names = makefile_parsed['SOURCES'].split()
		if lib_name.startswith('libmceliece') and 'aes256ctr.c' in _source_names:
			# Remove test infrastructure
			_source_names.remove('aes256ctr.c')
		sources = [(build_root / fn) for fn in _source_names]
	elif 'OBJECTS' in makefile_parsed.keys():
		_object_names = makefile_parsed['OBJECTS'].split()
		_objects = [(build_root / fn) for fn in _object_names]
		sources = [
		  next(p for p in extant_with_other_suffix(_p) if not p.suffix == '.h')
		  for _p in _objects
		]
	else:
		raise Exception(f"couldn't interpret Makefile while building {lib_name} from {build_root}")

	sources, extra_objects = partition_list(
	    lambda p: p.suffix == '.c',
	    sources
	)

	for fn in common_sources:
		##warnings.warn(f'FIXME: build-time inclusion of PQClean {fn} into {lib_name}')
		# https://stackoverflow.com/questions/77689317/dynamic-linking-in-python-cffi
		# https://github.com/python-cffi/cffi/issues/43
		# https://github.com/JamesTheAwesomeDude/pypqc/issues/1
		p = (common_dir / fn)
		sources.append(p)

	# 4. included_ffis, extra_compile_args, libraries, include_dirs #

	included_ffis = []

	extra_compile_args = []
	if platform.system() == 'Windows':
		# https://foss.heptapod.net/pypy/cffi/-/issues/516
		# https://www.reddit.com/r/learnpython/comments/175js2u/def_extern_says_im_not_using_it_in_api_mode/
		# https://learn.microsoft.com/en-us/cpp/build/reference/tc-tp-tc-tp-specify-source-file-type?view=msvc-170
		extra_compile_args.append('/TC')

	libraries = []
	if platform.system() == 'Windows':
		# https://stackoverflow.com/questions/69900013/link-error-cannot-build-python-c-extension-in-windows
		# https://learn.microsoft.com/en-us/windows/win32/seccrypto/required-libraries
		libraries.append('Advapi32')

	include_dirs = [(build_root), (common_dir)]

	# 5. create, return #

	ffibuilder = FFI()
	map_immed(ffibuilder.include, included_ffis)
	map_immed(ffibuilder.cdef, cdefs)
	ffibuilder.set_source(
		module_name,
		'\n'.join(c_header_sources),
		sources=[p.as_posix() for p in sources],
		include_dirs=[p.as_posix() for p in include_dirs],
		extra_objects=extra_objects,
		extra_compile_args=extra_compile_args,
		libraries=libraries,
	)
	return ffibuilder



from ._kem_cffi_maker import make_kem_ffi
from textwrap import dedent

def make_kyber_ffi(build_root):
	common_sources = ['fips202.c', 'randombytes.c']

	patent_info = (
		1,[
		'FR2956541A1/US9094189B2/EP2537284B1',
		'US9246675/EP2837128B1',
		'potential unknown others'], [
		'https://ntruprime.cr.yp.to/faq.html',
		'https://csrc.nist.gov/csrc/media/Projects/post-quantum-cryptography/documents/selected-algos-2022/nist-pqc-license-summary-and-excerpts.pdf',
		'https://groups.google.com/a/list.nist.gov/g/pqc-forum/c/G0DoD7lkGPk/m/d7Zw0qhGBwAJ',
		'https://datatracker.ietf.org/meeting/116/proceedings#pquip:~:text=Patents%20and%20PQC',
		'https://mailarchive.ietf.org/arch/msg/pqc/MS92cuZkSRCDEjpPP90s2uAcRPo/']
	)

	extra_cdefs = [dedent("""\
	// Exposed internal interface
	void %(namespace)sindcpa_enc(uint8_t *c, const uint8_t *m, const uint8_t *pk, const uint8_t *coins);
	void %(namespace)sindcpa_dec(uint8_t *m, const uint8_t *c, const uint8_t *sk);
	""")]

	extra_c_header_sources = [dedent("""\
	// Exposed internal interface
	#include "indcpa.h"
	""")]

	return make_kem_ffi(build_root=build_root, extra_c_header_sources=extra_c_header_sources, extra_cdefs=extra_cdefs, common_sources=common_sources, patent_info=patent_info)



from ._kem_cffi_maker import make_kem_ffi
from textwrap import dedent

def make_mceliece_ffi(build_root):
	common_sources = ['fips202.c', 'randombytes.c']

	extra_cdefs = [dedent("""\
	// Exposed internal interface
	typedef ... gf;
	int %(namespace)spk_gen(uint8_t *pk, uint8_t *sk, const uint32_t *perm, int16_t *pi, uint64_t *pivots);
	void %(namespace)sencrypt(uint8_t *s, const uint8_t *pk, uint8_t *e);
	int %(namespace)sdecrypt(uint8_t *e, const uint8_t *sk, const uint8_t *c);
	int %(namespace)sgenpoly_gen(gf *out, gf *f);
	#define SYS_N ...
	#define SYS_T ...
	#define GFBITS ...
	#define SYND_BYTES ...
	""")]

	extra_c_header_sources = [dedent("""\
	// Exposed internal interface
	#include "encrypt.h"
	#include "decrypt.h"
	#include "params.h"
	#include "sk_gen.h"
	""")]

	return make_kem_ffi(build_root=build_root, extra_c_header_sources=extra_c_header_sources, extra_cdefs=extra_cdefs, common_sources=common_sources)



from ._common_cffi_maker import make_pqclean_ffi
from textwrap import dedent

def make_sign_ffi(build_root, extra_c_header_sources=frozenset(), extra_cdefs=frozenset(), **k):
	cdefs = [dedent("""\
	// Public signature interface
	int %(namespace)scrypto_sign_keypair(uint8_t *pk, uint8_t *sk);
	int %(namespace)scrypto_sign_signature(uint8_t *sig, size_t *siglen, const uint8_t *m, size_t mlen, const uint8_t *sk);
	int %(namespace)scrypto_sign_verify(const uint8_t *sig, size_t siglen, const uint8_t *m, size_t mlen, const uint8_t *pk);
	int %(namespace)scrypto_sign(uint8_t *sm, size_t *smlen, const uint8_t *m, size_t mlen, const uint8_t *sk);
	int %(namespace)scrypto_sign_open(uint8_t *m, size_t *mlen, const uint8_t *sm, size_t smlen, const uint8_t *pk);
	#define %(namespace)sCRYPTO_SECRETKEYBYTES ...
	#define %(namespace)sCRYPTO_PUBLICKEYBYTES ...
	#define %(namespace)sCRYPTO_BYTES ...
	""")]

	c_header_sources = [dedent("""\
	// Public signature interface
	#include "api.h"
	""")]

	cdefs.append(dedent("""\
	// Site interface
	static const char _NAMESPACE[...];
	typedef uint8_t %(namespace)scrypto_secretkey[...];
	typedef uint8_t %(namespace)scrypto_publickey[...];
	typedef uint8_t %(namespace)scrypto_signature[...];
	"""))

	c_header_sources.append(dedent("""\
	// Site interface
	static const char _NAMESPACE[] = "%(namespace)s";
	typedef uint8_t %(namespace)scrypto_secretkey[%(namespace)sCRYPTO_SECRETKEYBYTES];
	typedef uint8_t %(namespace)scrypto_publickey[%(namespace)sCRYPTO_PUBLICKEYBYTES];
	typedef uint8_t %(namespace)scrypto_signature[%(namespace)sCRYPTO_BYTES];
	"""))

	cdefs.extend(extra_cdefs)
	c_header_sources.extend(extra_c_header_sources)

	return make_pqclean_ffi(build_root=build_root, c_header_sources=c_header_sources, cdefs=cdefs, **k)



from ._sign_cffi_maker import make_sign_ffi
from textwrap import dedent
import re

def make_falcon_ffi(build_root):
	common_sources = ['fips202.c', 'randombytes.c']

	patent_info = (
		2,
                ['US7308097B2'], [
		'https://csrc.nist.gov/csrc/media/Projects/post-quantum-cryptography/documents/selected-algos-2022/final-ip-statements/Falcon-Statements-final.pdf#page=20']
	)

	return make_sign_ffi(build_root=build_root, common_sources=common_sources, patent_info=patent_info)



from .._lib.libhqc_192_clean import ffi, lib

__all__ = ['kem_keypair', 'kem_enc', 'kem_dec']

_LIB_NAMESPACE = ffi.string(lib._NAMESPACE).decode('ascii')
_T_PUBLICKEY = f'{_LIB_NAMESPACE}crypto_publickey'
_T_SECRETKEY = f'{_LIB_NAMESPACE}crypto_secretkey'
_T_KEM_PLAINTEXT = f'{_LIB_NAMESPACE}crypto_kem_plaintext'
_T_KEM_CIPHERTEXT = f'{_LIB_NAMESPACE}crypto_kem_ciphertext'

_crypto_kem_keypair = getattr(lib, f'{_LIB_NAMESPACE}crypto_kem_keypair')
_crypto_kem_enc = getattr(lib, f'{_LIB_NAMESPACE}crypto_kem_enc')
_crypto_kem_dec = getattr(lib, f'{_LIB_NAMESPACE}crypto_kem_dec')


def keypair():
	pk = ffi.new(_T_PUBLICKEY)
	sk = ffi.new(_T_SECRETKEY)

	errno = _crypto_kem_keypair(ffi.cast('char*', pk), ffi.cast('char*', sk))

	if errno:
		raise RuntimeError(f"{_LIB_NAMESPACE}crypto_kem_keypair returned error code {errno}")
	return bytes(pk), bytes(sk)


def encap(pk):
	ciphertext = ffi.new(_T_KEM_CIPHERTEXT)
	key = ffi.new(_T_KEM_PLAINTEXT)
	pk = ffi.cast(_T_PUBLICKEY, ffi.from_buffer(pk))

	errno = _crypto_kem_enc(ffi.cast('char*', ciphertext), ffi.cast('char*', key), ffi.cast('char*', pk))

	if errno:
		raise RuntimeError(f"{_LIB_NAMESPACE}crypto_kem_enc returned error code {errno}")

	return bytes(key), bytes(ciphertext)


def decap(ciphertext, sk):
	key = ffi.new(_T_KEM_PLAINTEXT)
	ciphertext = ffi.cast(_T_KEM_CIPHERTEXT, ffi.from_buffer(ciphertext))
	sk = ffi.cast(_T_SECRETKEY, ffi.from_buffer(sk))

	errno = _crypto_kem_dec(ffi.cast('char*', key), ffi.cast('char*', ciphertext), ffi.cast('char*', sk))

	if errno:
		raise RuntimeError(f"{_LIB_NAMESPACE}crypto_kem_dec returned error code {errno}")

	return bytes(key)




from .._lib.libsphincs_shake_256f_simple_clean import ffi, lib

__all__ = ['keypair', 'signature', 'verify']

_LIB_NAMESPACE = ffi.string(lib._NAMESPACE).decode('ascii')
_T_PUBLICKEY = f'{_LIB_NAMESPACE}crypto_publickey'
_T_SECRETKEY = f'{_LIB_NAMESPACE}crypto_secretkey'
_T_SIGNATURE = f'{_LIB_NAMESPACE}crypto_signature'

_crypto_sign_keypair = getattr(lib, f'{_LIB_NAMESPACE}crypto_sign_keypair')
_crypto_sign_signature = getattr(lib, f'{_LIB_NAMESPACE}crypto_sign_signature')
_crypto_sign_verify = getattr(lib, f'{_LIB_NAMESPACE}crypto_sign_verify')


def keypair():
	_pk = ffi.new(_T_PUBLICKEY)
	_sk = ffi.new(_T_SECRETKEY)

	errno = _crypto_sign_keypair(_pk, _sk)

	if errno:
		raise RuntimeError(f"{_LIB_NAMESPACE}crypto_sign_keypair returned error code {errno}")
	return bytes(_pk), bytes(_sk)


def signature(m, sk):
	_m = ffi.from_buffer(m)

	_sk = ffi.cast(_T_SECRETKEY, ffi.from_buffer(sk))
	assert len(_sk) == len(sk)

	_sig = ffi.new(_T_SIGNATURE)
	_siglen = ffi.new('size_t*')

	errno = _crypto_sign_signature(_sig, _siglen, _m, len(m), _sk)
	assert len(_sig) == _siglen[0]  # This is a fixed parameter; WHY is it output??

	if errno:
		raise RuntimeError(f"{_LIB_NAMESPACE}crypto_sign_signature returned error code {errno}")

	return bytes(_sig)


def verify(sig, m, pk):
	_sig = ffi.cast(_T_SIGNATURE, ffi.from_buffer(sig))

	_m = ffi.from_buffer(m)

	_pk = ffi.cast(_T_PUBLICKEY, ffi.from_buffer(pk))

	errno = _crypto_sign_verify(_sig, len(_sig), _m, len(_m), _pk)

	if errno:
		if errno == -1:
			raise ValueError('verification failed')
		raise RuntimeError(f"{_LIB_NAMESPACE}crypto_sign_verify returned error code {errno}")

	return




# https://foss.heptapod.net/pypy/cffi/-/issues/441
# https://github.com/pypa/setuptools/issues/1040

from setuptools import setup

setup(
    cffi_modules=[
        'cffi_modules/dilithium2_clean.py:ffi',
        'cffi_modules/dilithium3_clean.py:ffi',
        'cffi_modules/dilithium5_clean.py:ffi',
        'cffi_modules/falcon_512_clean.py:ffi',
        'cffi_modules/falcon_1024_clean.py:ffi',
        'cffi_modules/hqc_128_clean.py:ffi',
        'cffi_modules/hqc_192_clean.py:ffi',
        'cffi_modules/hqc_256_clean.py:ffi',
        'cffi_modules/kyber512_clean.py:ffi',
        'cffi_modules/kyber768_clean.py:ffi',
        'cffi_modules/kyber1024_clean.py:ffi',
        'cffi_modules/mceliece348864f_clean.py:ffi',
        'cffi_modules/mceliece460896f_clean.py:ffi',
        'cffi_modules/mceliece6688128f_clean.py:ffi',
        'cffi_modules/mceliece6960119f_clean.py:ffi',
        'cffi_modules/mceliece8192128f_clean.py:ffi',
#        'cffi_modules/mceliece6688128pcf_clean.py:ffi',
#        'cffi_modules/mceliece6960119pcf_clean.py:ffi',
#        'cffi_modules/mceliece8192128pcf_clean.py:ffi',
        'cffi_modules/sphincs-sha2-128f-simple_clean.py:ffi',
        'cffi_modules/sphincs-sha2-128s-simple_clean.py:ffi',
        'cffi_modules/sphincs-sha2-192f-simple_clean.py:ffi',
        'cffi_modules/sphincs-sha2-192s-simple_clean.py:ffi',
        'cffi_modules/sphincs-sha2-256f-simple_clean.py:ffi',
        'cffi_modules/sphincs-sha2-256s-simple_clean.py:ffi',
        'cffi_modules/sphincs-shake-128f-simple_clean.py:ffi',
        'cffi_modules/sphincs-shake-128s-simple_clean.py:ffi',
        'cffi_modules/sphincs-shake-192f-simple_clean.py:ffi',
        'cffi_modules/sphincs-shake-192s-simple_clean.py:ffi',
        'cffi_modules/sphincs-shake-256f-simple_clean.py:ffi',
        'cffi_modules/sphincs-shake-256s-simple_clean.py:ffi',
    ],
)



from ._sign_cffi_maker import make_sign_ffi
from textwrap import dedent

def make_dilithium_ffi(build_root):
	common_sources = ['fips202.c', 'randombytes.c']

	extra_cdefs = [dedent("""\
	// Exposed internal interface
	#define CRHBYTES ...
	#define RNDBYTES ...
	#define N ...
	#define Q ...
	#define D ...
	#define ROOT_OF_UNITY ...
	#define K ...
	#define L ...
	#define ETA ...
	#define TAU ...
	#define BETA ...
	#define GAMMA1 ...
	#define GAMMA2 ...
	#define OMEGA ...
	#define CTILDEBYTES ...
	""")]

	extra_c_header_sources = [dedent("""\
	// Exposed internal interface
	#include "params.h"
	""")]

	return make_sign_ffi(build_root=build_root, extra_c_header_sources=extra_c_header_sources, extra_cdefs=extra_cdefs, common_sources=common_sources)



from collections import deque
from functools import partial
from itertools import starmap
from pathlib import Path
import platform
import re
from textwrap import dedent
from warnings import warn

def extant_with_other_suffix(p):
	assert not re.match(r'[\?\*\[]', p.stem)
	pseudo_p = p.with_suffix('.*')
	return pseudo_p.parent.glob(pseudo_p.name)


def using_avx2():
	return False  # TODO


def partition_list(predicate, it):
	l_true = []
	l_false = []
	for item in it:
		if predicate(item):
			l_true.append(item)
		else:
			l_false.append(item)
	return l_true, l_false


def map_immed(f, it, *, splat=False):
	deque((map if not splat else starmap)(f, it), 0)


def patent_warning(subject, patent_info):
	severity, patents, links = patent_info

	if severity == 0:
		return None

	if severity == 1:
		return dedent(f"""\
			{subject} may be protected under patent(s) {'; '.join(patents)}.
			If you rely on this library via PyPI, it could break at any time if I'm forced by the patentholders to remove this module.
			Additionally, the patentholders might impose on you *additional* terms, beyond those stated in the software's license.
			
			This is not legal advice. For more information, see:
			""") + '\n'.join(links) + dedent(f"""
			
			If the continued use of {subject} is important to you, consider hiring a lawyer and/or purchasing a license for it.""")

	if severity == 2:
		return dedent(f"""\
			{subject} may be protected under patent(s) {'; '.join(patents)}.
			ITS LICENSING STATUS FOR PUBLIC USE IS DISPUTED OR UNKNOWN AT THIS TIME.
			If you rely on this library via PyPI, it could break at any time if I'm forced by the patentholders to remove this module.
			Additionally, the patentholders might impose on you *additional* terms, beyond those stated in the software's license.
			
			This is not legal advice. For more information, see:
			""") + '\n'.join(links) + dedent(f"""
			
			If the continued use of {subject} is important to you, consider hiring a lawyer and/or purchasing a license for it.""")

	if severity == 3:
		return dedent(f"""\
			{subject} may be protected under patent(s) {'; '.join(patents)}.
			IT MIGHT NOT BE LICENSED FOR PUBLIC USE AT THIS TIME.
			If you rely on this library via PyPI, it could break at any time if I'm forced by the patentholders to remove this module.
			Additionally, the patentholders might impose on you *additional* terms, beyond those stated in the software's license.
			
			This is not legal advice. For more information, see:
			""") + '\n'.join(links) + dedent(f"""
			
			If the continued use of {subject} is important to you, consider hiring a lawyer and/or purchasing a license for it.""")

	raise ValueError(f'severity = {severity}')
